From 2112c82c260dbe731df4619cd17c67e2ff280cf8 Mon Sep 17 00:00:00 2001
From: David Fletcher <david@megapico.co.uk>
Date: Thu, 5 May 2022 00:07:07 +0100
Subject: [PATCH] Patched to work with libmad MP3 decoding codec (update 2)

Signed-off-by: David Fletcher <david@megapico.co.uk>
---
 configure              |   4 +
 libavcodec/Makefile    |   1 +
 libavcodec/allcodecs.c |   1 +
 libavcodec/libmaddec.c | 174 +++++++++++++++++++++++++++++++++++++++++
 4 files changed, 180 insertions(+)
 create mode 100644 libavcodec/libmaddec.c

diff --git a/configure b/configure
index 196873c4aa..d5f5b79614 100755
--- a/configure
+++ b/configure
@@ -245,6 +245,7 @@ External library support:
   --enable-libklvanc       enable Kernel Labs VANC processing [no]
   --enable-libkvazaar      enable HEVC encoding via libkvazaar [no]
   --enable-liblensfun      enable lensfun lens correction [no]
+  --enable-libmad          enable MP3 decoding via libmad [no]
   --enable-libmodplug      enable ModPlug via libmodplug [no]
   --enable-libmp3lame      enable MP3 encoding via libmp3lame [no]
   --enable-libopencore-amrnb enable AMR-NB de/encoding via libopencore-amrnb [no]
@@ -1775,6 +1776,7 @@ EXTERNAL_LIBRARY_GPL_LIST="
     frei0r
     libcdio
     libdavs2
+    libmad
     librubberband
     libvidstab
     libx264
@@ -3339,6 +3341,7 @@ libjxl_decoder_deps="libjxl libjxl_threads"
 libjxl_encoder_deps="libjxl libjxl_threads"
 libkvazaar_encoder_deps="libkvazaar"
 libmodplug_demuxer_deps="libmodplug"
+libmad_decoder_deps="libmad"
 libmp3lame_encoder_deps="libmp3lame"
 libmp3lame_encoder_select="audio_frame_queue mpegaudioheader"
 libopencore_amrnb_decoder_deps="libopencore_amrnb"
@@ -6572,6 +6575,7 @@ if enabled libmfx; then
 fi
 
 enabled libmodplug        && require_pkg_config libmodplug libmodplug libmodplug/modplug.h ModPlug_Load
+enabled libmad            && require libmad "mad.h" mad_decoder_init -lmad
 enabled libmp3lame        && require "libmp3lame >= 3.98.3" lame/lame.h lame_set_VBR_quality -lmp3lame $libm_extralibs
 enabled libmysofa         && { check_pkg_config libmysofa libmysofa mysofa.h mysofa_neighborhood_init_withstepdefine ||
                                require libmysofa mysofa.h mysofa_neighborhood_init_withstepdefine -lmysofa $zlib_extralibs; }
diff --git a/libavcodec/Makefile b/libavcodec/Makefile
index cfaa6f196a..1e434222d9 100644
--- a/libavcodec/Makefile
+++ b/libavcodec/Makefile
@@ -1065,6 +1065,7 @@ OBJS-$(CONFIG_LIBILBC_ENCODER)            += libilbc.o
 OBJS-$(CONFIG_LIBJXL_DECODER)             += libjxldec.o libjxl.o
 OBJS-$(CONFIG_LIBJXL_ENCODER)             += libjxlenc.o libjxl.o
 OBJS-$(CONFIG_LIBKVAZAAR_ENCODER)         += libkvazaar.o
+OBJS-$(CONFIG_LIBMAD_DECODER)             += libmaddec.o
 OBJS-$(CONFIG_LIBMP3LAME_ENCODER)         += libmp3lame.o
 OBJS-$(CONFIG_LIBOPENCORE_AMRNB_DECODER)  += libopencore-amr.o
 OBJS-$(CONFIG_LIBOPENCORE_AMRNB_ENCODER)  += libopencore-amr.o
diff --git a/libavcodec/allcodecs.c b/libavcodec/allcodecs.c
index c47133aa18..0b8430db92 100644
--- a/libavcodec/allcodecs.c
+++ b/libavcodec/allcodecs.c
@@ -744,6 +744,7 @@ extern const FFCodec ff_libcodec2_decoder;
 extern const FFCodec ff_libdav1d_decoder;
 extern const FFCodec ff_libdavs2_decoder;
 extern const FFCodec ff_libfdk_aac_encoder;
+extern const FFCodec ff_libmad_decoder;
 extern const FFCodec ff_libfdk_aac_decoder;
 extern const FFCodec ff_libgsm_encoder;
 extern const FFCodec ff_libgsm_decoder;
diff --git a/libavcodec/libmaddec.c b/libavcodec/libmaddec.c
new file mode 100644
index 0000000000..4812c5e119
--- /dev/null
+++ b/libavcodec/libmaddec.c
@@ -0,0 +1,174 @@
+/*
+ * MP3 decoder using libmad
+ * Copyright (c) 2022 David Fletcher
+ *
+ * This file is part of FFmpeg.
+ *
+ * FFmpeg is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ *
+ * FFmpeg is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with FFmpeg; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
+ */
+
+#include <mad.h>
+
+#include "libavutil/channel_layout.h"
+#include "libavutil/common.h"
+#include "avcodec.h"
+#include "codec_internal.h"
+#include "internal.h"
+
+#define MAD_BUFSIZE (32 * 1024)
+#define MIN(a, b) ((a) < (b) ? (a) : (b))
+
+typedef struct libmad_context {
+    uint8_t input_buffer[MAD_BUFSIZE+MAD_BUFFER_GUARD];
+    struct mad_synth  synth; 
+    struct mad_stream stream;
+    struct mad_frame  frame;
+    struct mad_header header;
+    int got_header;
+}libmad_context;		
+
+/* utility to scale and round samples to 16 bits */
+static inline signed int mad_scale(mad_fixed_t sample)
+{
+     /* round */
+     sample += (1L << (MAD_F_FRACBITS - 16));
+ 
+     /* clip */
+     if (sample >= MAD_F_ONE)
+         sample = MAD_F_ONE - 1;
+     else if (sample < -MAD_F_ONE)
+         sample = -MAD_F_ONE;
+    
+     /* quantize */
+     return sample >> (MAD_F_FRACBITS + 1 - 16);
+}
+
+static av_cold int libmad_decode_init(AVCodecContext *avc)
+{
+     libmad_context *mad = avc->priv_data;
+
+     mad_synth_init  (&mad->synth);
+     mad_stream_init (&mad->stream);
+     mad_frame_init  (&mad->frame);
+     mad->got_header = 0;
+
+     return 0;
+}
+
+static av_cold int libmad_decode_close(AVCodecContext *avc)
+{
+     libmad_context *mad = avc->priv_data;
+
+     mad_synth_finish(&mad->synth);
+     mad_frame_finish(&mad->frame);
+     mad_stream_finish(&mad->stream);
+    
+     return 0;
+}
+
+static int libmad_decode_frame(AVCodecContext *avc, AVFrame *frame,
+                          int *got_frame_ptr, AVPacket *pkt)
+{
+     libmad_context *mad = avc->priv_data;
+     struct mad_pcm *pcm;
+     mad_fixed_t const *left_ch;
+     mad_fixed_t const *right_ch;
+     int16_t *output;
+     int nsamples;
+     int nchannels;
+     int ret;
+     size_t bytes_read = 0;
+     size_t remaining = 0;
+
+     remaining = mad->stream.bufend - mad->stream.next_frame;
+     memmove(mad->input_buffer, mad->stream.next_frame, remaining);
+     bytes_read = MIN(pkt->size, MAD_BUFSIZE - remaining);
+     memcpy(mad->input_buffer+remaining, pkt->data, bytes_read);
+     
+     if (bytes_read == 0){
+	 *got_frame_ptr = 0;
+	 return 0;
+     }
+     
+     mad_stream_buffer(&mad->stream, mad->input_buffer, remaining + bytes_read);
+     mad->stream.error = 0;
+     
+     if(!mad->got_header){
+	 mad_header_decode(&mad->header, &mad->stream);
+	 mad->got_header = 1;
+	 avc->frame_size = 32 * (mad->header.layer == MAD_LAYER_I ? 12 : \
+				 ((mad->header.layer == MAD_LAYER_III && \
+				   (mad->header.flags & MAD_FLAG_LSF_EXT)) ? 18 : 36));
+	 avc->sample_fmt = AV_SAMPLE_FMT_S16;
+         av_channel_layout_uninit(&avc->ch_layout);
+	 if(mad->header.mode == MAD_MODE_SINGLE_CHANNEL){
+             avc->ch_layout = (AVChannelLayout)AV_CHANNEL_LAYOUT_MONO;
+             avc->ch_layout.nb_channels = 1;
+	 }else{
+	     avc->ch_layout = (AVChannelLayout)AV_CHANNEL_LAYOUT_STEREO;
+             avc->ch_layout.nb_channels = 2;
+	 }
+     }
+     
+     frame->ch_layout = avc->ch_layout;
+     frame->format = avc->sample_fmt;
+     frame->ch_layout.nb_channels = avc->ch_layout.nb_channels;
+     frame->nb_samples = avc->frame_size; 
+     
+     if ((ret = ff_get_buffer(avc, frame, 0)) < 0)
+	 return ret;
+     
+     if (mad_frame_decode(&mad->frame, &mad->stream) == -1) {
+	 *got_frame_ptr = 0;
+	 return mad->stream.bufend - mad->stream.next_frame;
+     }
+     
+     mad_synth_frame (&mad->synth, &mad->frame);
+     
+     pcm = &mad->synth.pcm;
+     output = (int16_t *)frame->data[0];
+     nsamples = pcm->length;
+     nchannels = pcm->channels;
+     left_ch = pcm->samples[0];
+     right_ch = pcm->samples[1];
+     while (nsamples--) {
+	 *output++ = mad_scale(*(left_ch++));
+	 if (nchannels == 2) {
+	     *output++ = mad_scale(*(right_ch++));
+	 }
+	 //Players should recognise mono and play through both channels
+	 //Writing the same thing to both left and right channels here causes
+	 //memory issues as it creates double the number of samples allocated.
+     }
+     
+     *got_frame_ptr = 1;
+     
+     return mad->stream.bufend - mad->stream.next_frame;
+}
+
+const FFCodec ff_libmad_decoder = {
+    .p.name           = "libmad",
+    .p.long_name      = NULL_IF_CONFIG_SMALL("libmad MP3 decoder"),
+    .p.wrapper_name   = "libmad",
+    .p.type           = AVMEDIA_TYPE_AUDIO,
+    .p.id             = AV_CODEC_ID_MP3,
+    .p.sample_fmts    = (const enum AVSampleFormat[]) { AV_SAMPLE_FMT_S16, AV_SAMPLE_FMT_NONE },
+    .p.capabilities   = AV_CODEC_CAP_DR1 | AV_CODEC_CAP_CHANNEL_CONF,
+    .priv_data_size = sizeof(libmad_context),
+    .init           = libmad_decode_init,
+    .close          = libmad_decode_close,
+    FF_CODEC_DECODE_CB(libmad_decode_frame)
+};
+
-- 
2.17.6

